import pytest
from aio_pika import Message
from aio_pika.abc import AbstractRobustConnection

from khnm.pipes import get_queue_name
from khnm.producers import make_producer
from tests.doubles import RaisingMessageSender
from tests.utils import get_queue_messages_count


async def test_producing_message_adds_it_to_queue(
    amqp_connection: AbstractRobustConnection,
    sample_message: Message,
    pipe: str = "test",
) -> None:
    async with make_producer(amqp_connection, pipe) as producer:
        await producer.publish(sample_message)
    messages_count = await get_queue_messages_count(
        amqp_connection, get_queue_name(pipe)
    )
    assert messages_count == 1


async def test_producer_channel_is_closed_properly_on_exception(
    amqp_connection: AbstractRobustConnection,
    sample_message: Message,
    pipe: str = "test",
) -> None:
    async with make_producer(
        amqp_connection, pipe, sender=RaisingMessageSender()
    ) as producer:
        with pytest.raises(Exception):
            await producer.publish(sample_message)
    assert producer.is_closed



import argparse
import asyncio
import importlib
import os

from aio_pika import connect_robust

from khnm.pipelines import Pipeline


def load_pipeline_object(import_string: str) -> Pipeline:
    module_name, attribute_name = import_string.split(":", 1)
    module = importlib.import_module(module_name)
    pipeline = getattr(module, attribute_name)
    return pipeline


async def run_node(
    pipeline: Pipeline,
    node_name: str,
    connection_string: str,
    tasks: int = 1,
    threads: int = 1,
) -> None:
    connection = await connect_robust(connection_string)
    await asyncio.gather(
        *[pipeline.run(connection, node_name, threads) for _ in range(tasks)]
    )


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "pipeline",
        help="Where to look for pipeline object, in format 'module:attribute' (e.g. 'main:pipeline')",
    )
    parser.add_argument("node", help="Node name to run")
    parser.add_argument(
        "--tasks",
        type=int,
        default=1,
        help="Number of concurrent async tasks to run",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=1,
        help="Number of concurrent threads to run (sync callbacks only)",
    )
    args = parser.parse_args()

    conn_string = os.getenv("RABBITMQ_URL")
    if not conn_string:
        raise SystemExit("RABBITMQ_URL not set")
    if args.tasks < 1:
        raise SystemExit("Number of tasks must be greater than 0")
    if args.threads < 1:
        raise SystemExit("Number of threads must be greater than 0")

    pipeline = load_pipeline_object(args.pipeline)
    asyncio.run(run_node(pipeline, args.node, conn_string, args.tasks, args.threads))



import asyncio
import datetime
from typing import Protocol


class Clock(Protocol):
    def now(self) -> datetime.datetime: ...

    async def sleep(self, time_seconds: float) -> None: ...


class UtcClock(Clock):
    def now(self) -> datetime.datetime:
        return datetime.datetime.now(datetime.timezone.utc)

    async def sleep(self, time_seconds: float) -> None:
        await asyncio.sleep(time_seconds)



from khnm.serialization import pydantic_model_to_message, message_to_pydantic_model
from tests.doubles import SampleDataObject


async def test_string_is_serialized_correctly(
    data_object: SampleDataObject = SampleDataObject(),
) -> None:
    message = pydantic_model_to_message(data_object)
    deserialized_obj = message_to_pydantic_model(message, SampleDataObject)
    assert deserialized_obj.text == data_object.text


async def test_uuid_is_serialized_correctly(
    data_object: SampleDataObject = SampleDataObject(),
) -> None:
    message = pydantic_model_to_message(data_object)
    deserialized_obj = message_to_pydantic_model(message, SampleDataObject)
    assert deserialized_obj.guid == data_object.guid


async def test_time_is_serialized_correctly(
    data_object: SampleDataObject = SampleDataObject(),
) -> None:
    message = pydantic_model_to_message(data_object)
    deserialized_obj = message_to_pydantic_model(message, SampleDataObject)
    assert deserialized_obj.time == data_object.time


async def test_integer_is_serialized_correctly(
    data_object: SampleDataObject = SampleDataObject(),
) -> None:
    message = pydantic_model_to_message(data_object)
    deserialized_obj = message_to_pydantic_model(message, SampleDataObject)
    assert deserialized_obj.integer == data_object.integer


async def test_float_is_serialized_correctly(
    data_object: SampleDataObject = SampleDataObject(),
) -> None:
    message = pydantic_model_to_message(data_object)
    deserialized_obj = message_to_pydantic_model(message, SampleDataObject)
    assert deserialized_obj.float_ == data_object.float_


async def test_decimal_is_serialized_correctly(
    data_object: SampleDataObject = SampleDataObject(),
) -> None:
    message = pydantic_model_to_message(data_object)
    deserialized_obj = message_to_pydantic_model(message, SampleDataObject)
    assert deserialized_obj.decimal_ == data_object.decimal_


async def test_nested_object_is_serialized_correctly(
    data_object: SampleDataObject = SampleDataObject(),
) -> None:
    message = pydantic_model_to_message(data_object)
    deserialized_obj = message_to_pydantic_model(message, SampleDataObject)
    assert deserialized_obj.obj.text == data_object.obj.text



from contextlib import asynccontextmanager
from typing import Optional, Protocol, AsyncGenerator

from aio_pika import Message
from aio_pika.abc import AbstractRobustConnection, AbstractChannel

from khnm.pipes import declare_pipe, send_with_backoff, send_message
from khnm.time import Clock, UtcClock
from khnm.types import SenderT


class Producer(Protocol):
    async def publish(self, message: Message) -> None: ...
    @property
    def is_closed(self) -> bool: ...


class AmqpProducer(Producer):
    def __init__(
        self,
        channel: AbstractChannel,
        pipe: str,
        backoff_seconds: float = 0.1,
        max_retries: Optional[int] = None,
        exponential_backoff: bool = False,
        max_backoff_seconds: Optional[float] = None,
        apply_jitter: bool = False,
        persistent_messages: bool = False,
        clock: Clock = UtcClock(),
        sender: SenderT = send_message,
    ) -> None:
        self._channel = channel
        self._pipe = pipe
        self._backoff_seconds = backoff_seconds
        self._max_retries = max_retries
        self._exponential_backoff = exponential_backoff
        self._max_backoff_seconds = max_backoff_seconds
        self._apply_jitter = apply_jitter
        self._persistent_messages = persistent_messages
        self._clock = clock
        self._sender = sender

    async def publish(self, message: Message) -> None:
        await send_with_backoff(
            channel=self._channel,
            sender=self._sender,
            message=message,
            pipe=self._pipe,
            backoff_seconds=self._backoff_seconds,
            max_retries=self._max_retries,
            exponential_backoff=self._exponential_backoff,
            max_backoff_seconds=self._max_backoff_seconds,
            apply_jitter=self._apply_jitter,
            persistent=self._persistent_messages,
            clock=self._clock,
        )

    @property
    def is_closed(self) -> bool:
        return self._channel.is_closed


@asynccontextmanager
async def make_producer(
    amqp_connection: AbstractRobustConnection,
    pipe: str,
    size: Optional[int] = None,
    durable: bool = False,
    backoff_seconds: float = 0.1,
    max_retries: Optional[int] = None,
    exponential_backoff: bool = False,
    max_backoff_seconds: Optional[float] = None,
    apply_jitter: bool = False,
    persistent_messages: bool = False,
    clock: Clock = UtcClock(),
    sender: SenderT = send_message,
) -> AsyncGenerator[Producer, None]:
    channel = await amqp_connection.channel()
    await declare_pipe(channel=channel, name=pipe, size=size, durable=durable)
    try:
        yield AmqpProducer(
            channel=channel,
            pipe=pipe,
            backoff_seconds=backoff_seconds,
            max_retries=max_retries,
            exponential_backoff=exponential_backoff,
            max_backoff_seconds=max_backoff_seconds,
            apply_jitter=apply_jitter,
            persistent_messages=persistent_messages,
            clock=clock,
            sender=sender,
        )
    finally:
        await channel.close()



from typing import Optional

import pytest
from aio_pika import Message
from aio_pika.abc import AbstractRobustConnection

from khnm.time import Clock
from khnm.pipes import declare_pipe, send_message, send_with_backoff, wait_for_pipe
from khnm.types import QueueGetterT
from tests.doubles import FailingMessageSender, FakeClock, FailingQueueGetter


@pytest.mark.parametrize("pipe_size", [None, 1, 4, 1024])
async def test_message_is_published_when_pipe_max_size_not_exceeded(
    amqp_connection: AbstractRobustConnection,
    sample_message: Message,
    pipe_size: Optional[int],
    pipe: str = "test",
) -> None:
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe, pipe_size)
        message_sent = await send_message(channel, sample_message, pipe)
    assert message_sent


@pytest.mark.parametrize("pipe_size", [0, 1, 4, 1024])
async def test_message_is_rejected_if_pipe_max_size_exceeded(
    amqp_connection: AbstractRobustConnection,
    sample_message: Message,
    pipe_size: int,
    pipe: str = "test",
) -> None:
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe, pipe_size)
        for _ in range(pipe_size):
            await send_message(channel, sample_message, pipe)
        overflowing_message_sent = await send_message(channel, sample_message, pipe)
    assert not overflowing_message_sent


@pytest.mark.parametrize("max_retries", [0, 1, 3, 1024])
async def test_message_is_retried_until_max_retries_exceeded(
    amqp_connection: AbstractRobustConnection,
    clock: Clock,
    sample_message: Message,
    max_retries: int,
    backoff_seconds: float = 1.0,
    pipe: str = "test",
) -> None:
    start_time = clock.now()
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe, size=0)
        await send_with_backoff(
            channel,
            send_message,
            sample_message,
            pipe,
            backoff_seconds=backoff_seconds,
            max_retries=max_retries,
            clock=clock,
        )
    end_time = clock.now()
    time_delta = end_time - start_time
    assert time_delta.total_seconds() == backoff_seconds * max_retries


async def test_zero_retries_does_not_prevent_sending(
    amqp_connection: AbstractRobustConnection,
    clock: Clock,
    sample_message: Message,
    max_retries: int = 0,
    pipe: str = "test",
) -> None:
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe)
        success = await send_with_backoff(
            channel,
            send_message,
            sample_message,
            pipe,
            max_retries=max_retries,
            clock=clock,
        )
    assert success


async def test_no_max_retries_allow_infinite_backoff(
    amqp_connection: AbstractRobustConnection,
    clock: Clock,
    sample_message: Message,
    sender=FailingMessageSender(fails_count=999999),
    pipe: str = "test",
) -> None:
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe)
        success = await send_with_backoff(
            channel, sender, sample_message, pipe, max_retries=None, clock=clock
        )
    assert success


@pytest.mark.parametrize(
    "max_retries,expected_backoff_sum",
    [
        (1, 1),
        (2, 3),
        (3, 7),
        (4, 15),
        (5, 31),
        (6, 63),
        (7, 127),
        (8, 255),
        (9, 511),
        (10, 1023),
    ],
)
async def test_exponential_backoff_waits_correct_times(
    amqp_connection: AbstractRobustConnection,
    clock: Clock,
    sample_message: Message,
    max_retries: int,
    expected_backoff_sum: float,
    initial_backoff_seconds: float = 1.0,
    pipe: str = "test",
) -> None:
    start_time = clock.now()
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe, size=0)
        await send_with_backoff(
            channel,
            send_message,
            sample_message,
            pipe,
            backoff_seconds=initial_backoff_seconds,
            max_retries=max_retries,
            exponential_backoff=True,
            clock=clock,
        )
    end_time = clock.now()
    time_delta = end_time - start_time
    assert time_delta.total_seconds() == expected_backoff_sum


@pytest.mark.parametrize(
    "max_retries,expected_backoff_sum",
    [
        (1, 1),
        (2, 3),
        (3, 7),
        (4, 15),
        (5, 31),
        (6, 63),
        (7, 123),
        (8, 183),
        (9, 243),
        (10, 303),
    ],
)
async def test_max_backoff_cuts_out_exponential_backoff_time(
    amqp_connection: AbstractRobustConnection,
    clock: Clock,
    sample_message: Message,
    max_retries: int,
    expected_backoff_sum: float,
    initial_backoff_seconds: float = 1.0,
    max_backoff_seconds: float = 60,
    pipe: str = "test",
) -> None:
    start_time = clock.now()
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe, size=0)
        await send_with_backoff(
            channel,
            send_message,
            sample_message,
            pipe,
            backoff_seconds=initial_backoff_seconds,
            max_retries=max_retries,
            exponential_backoff=True,
            max_backoff_seconds=max_backoff_seconds,
            clock=clock,
        )
    end_time = clock.now()
    time_delta = end_time - start_time
    assert time_delta.total_seconds() == expected_backoff_sum


@pytest.mark.parametrize(
    "max_retries,initial_backoff_seconds",
    [
        (1, 1),
        (2, 3),
        (3, 7),
        (4, 15),
        (5, 31),
        (6, 63),
        (7, 123),
        (8, 183),
        (9, 243),
        (10, 303),
    ],
)
async def test_jitter_randomizes_backoff_times(
    amqp_connection: AbstractRobustConnection,
    clock: Clock,
    sample_message: Message,
    max_retries: int,
    initial_backoff_seconds: float,
    pipe: str = "test",
) -> None:
    clock_1 = FakeClock(clock.now())
    clock_2 = FakeClock(clock.now())

    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe, size=0)

        await send_with_backoff(
            channel,
            send_message,
            sample_message,
            pipe,
            backoff_seconds=initial_backoff_seconds,
            exponential_backoff=True,
            max_retries=max_retries,
            apply_jitter=True,
            clock=clock_1,
        )

        await send_with_backoff(
            channel,
            send_message,
            sample_message,
            pipe,
            backoff_seconds=initial_backoff_seconds,
            exponential_backoff=True,
            max_retries=max_retries,
            apply_jitter=True,
            clock=clock_2,
        )

    assert clock_1.time != clock_2.time


async def test_waiting_for_pipe_returns_true_when_queue_exists(
    amqp_connection: AbstractRobustConnection, pipe: str = "test"
) -> None:
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe, size=0)
        success = await wait_for_pipe(channel, pipe, max_retries=0)
    assert success


async def test_waiting_for_pipe_fails_if_pipe_not_declared(
    amqp_connection: AbstractRobustConnection, pipe: str = "test", max_retries: int = 0
) -> None:
    async with amqp_connection.channel() as channel:
        success = await wait_for_pipe(channel, pipe, max_retries=max_retries)
    assert not success


@pytest.mark.parametrize("max_retries", [0, 1, 3, 4, 16, 32, 64, 1024])
async def test_waiting_for_pipe_retries_until_max_retries(
    amqp_connection: AbstractRobustConnection,
    max_retries: int,
    clock: Clock,
    queue_getter: QueueGetterT = FailingQueueGetter(9999),
    pipe: str = "test",
    backoff_seconds: float = 1.0,
) -> None:
    start_time = clock.now()
    async with amqp_connection.channel() as channel:
        await wait_for_pipe(
            channel,
            pipe,
            backoff_seconds=backoff_seconds,
            max_retries=max_retries,
            clock=clock,
            getter=queue_getter,
        )
    end_time = clock.now()
    time_delta = end_time - start_time
    seconds_passed = time_delta.total_seconds()
    assert seconds_passed == backoff_seconds * max_retries


async def test_if_no_max_retries_pipe_is_waited_for_infinitely(
    amqp_connection: AbstractRobustConnection,
    clock: Clock,
    queue_getter: QueueGetterT = FailingQueueGetter(9999),
    pipe: str = "test",
) -> None:
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe, size=0)
        success = await wait_for_pipe(
            channel, pipe, clock=clock, max_retries=None, getter=queue_getter
        )
    assert success


async def test_backoff_is_respected_when_retries_set_infinite(
    amqp_connection: AbstractRobustConnection,
    clock: Clock,
    sample_message: Message,
    pipe: str = "test",
) -> None:
    fails_count = 10
    start_time = clock.now()
    sender = FailingMessageSender(fails_count)
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe)
        await send_with_backoff(
            channel,
            sender,
            sample_message,
            pipe,
            backoff_seconds=1,
            exponential_backoff=False,
            max_retries=None,
            apply_jitter=False,
            clock=clock,
        )
    stop_time = clock.now()
    delta = stop_time - start_time
    seconds_passed = delta.total_seconds()
    assert seconds_passed == fails_count


async def test_exponential_backoff_is_respected_when_retries_set_infinite(
    amqp_connection: AbstractRobustConnection,
    clock: Clock,
    sample_message: Message,
    pipe: str = "test",
) -> None:
    fails_count = 10
    start_time = clock.now()
    sender = FailingMessageSender(fails_count)
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe)
        await send_with_backoff(
            channel,
            sender,
            sample_message,
            pipe,
            backoff_seconds=1,
            exponential_backoff=True,
            max_retries=None,
            apply_jitter=False,
            clock=clock,
        )
    stop_time = clock.now()
    delta = stop_time - start_time
    seconds_passed = delta.total_seconds()
    assert seconds_passed == 2**10 - 1


async def test_infinite_retries_applies_jitter(
    amqp_connection: AbstractRobustConnection,
    sample_message: Message,
    clock: Clock,
    pipe: str = "test",
) -> None:
    fails_count = 10
    clock_1 = FakeClock(clock.now())
    clock_2 = FakeClock(clock_1.now())
    sender_1 = FailingMessageSender(fails_count)
    sender_2 = FailingMessageSender(fails_count)

    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe)
        await send_with_backoff(
            channel,
            sender_1,
            sample_message,
            pipe,
            backoff_seconds=1,
            exponential_backoff=False,
            max_retries=None,
            apply_jitter=True,
            clock=clock_1,
        )
        await send_with_backoff(
            channel,
            sender_2,
            sample_message,
            pipe,
            backoff_seconds=1,
            exponential_backoff=False,
            max_retries=None,
            apply_jitter=True,
            clock=clock_2,
        )
    assert clock_1.now() != clock_2.now()


async def test_max_backoff_cap_works_with_infinite_retries(
    amqp_connection: AbstractRobustConnection,
    sample_message: Message,
    clock: Clock,
    pipe: str = "test",
) -> None:
    fails_count = 10
    backoff_seconds_cap = 10

    start_time = clock.now()
    sender = FailingMessageSender(fails_count)
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe)
        await send_with_backoff(
            channel,
            sender,
            sample_message,
            pipe,
            backoff_seconds=1,
            exponential_backoff=True,
            max_retries=None,
            apply_jitter=False,
            max_backoff_seconds=backoff_seconds_cap,
            clock=clock,
        )
    stop_time = clock.now()
    delta = stop_time - start_time
    seconds_passed = delta.total_seconds()
    assert seconds_passed <= backoff_seconds_cap * fails_count



import asyncio
import inspect
from asyncio import AbstractEventLoop
from collections.abc import Iterable
from concurrent.futures.thread import ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import (
    Self,
    Callable,
    Protocol,
    List,
    Union,
    AsyncGenerator,
    cast,
    Generator,
    Awaitable,
    Optional,
    TypedDict,
    Unpack,
    AsyncContextManager,
    get_type_hints,
)

import pydantic
from aio_pika.abc import AbstractRobustConnection, AbstractIncomingMessage
from pydantic import BaseModel

from khnm.consumers import consume
from khnm.exceptions import NodeKwargsInvalid, PipelineDefinitionInvalid
from khnm.producers import make_producer, Producer
from khnm.serialization import pydantic_model_to_message, message_to_pydantic_model
from khnm.time import Clock, UtcClock
from khnm.types import GeneratorCallbackT, CallbackT, CallbackOutputT


class Bag(pydantic.BaseModel):
    model_config = {"extra": "allow"}


class Runner(Protocol):
    @property
    def name(self) -> str:
        pass

    async def run(self, connection: AbstractRobustConnection, threads: int) -> None:
        pass


class Source(Runner):
    def __init__(
        self,
        name: str,
        downstream_pipe: str,
        callback: GeneratorCallbackT,
        pipe_length: Optional[int] = None,
        durable: bool = False,
        backoff_seconds: float = 0.1,
        max_retries: Optional[int] = None,
        exponential_backoff: bool = False,
        max_backoff_seconds: Optional[float] = None,
        apply_jitter: bool = False,
        clock: Clock = UtcClock(),
    ) -> None:
        self._name = name
        self._downstream_pipe = downstream_pipe
        self._callback = callback
        self._pipe_length = pipe_length
        self._durable = durable
        self._backoff_seconds = backoff_seconds
        self._max_retries = max_retries
        self._exponential_backoff = exponential_backoff
        self._max_backoff_seconds = max_backoff_seconds
        self._apply_jitter = apply_jitter
        self._clock = clock

    @property
    def name(self) -> str:
        return self._name

    async def run(self, connection: AbstractRobustConnection, threads: int) -> None:
        is_callback_async = _is_callback_async(self._callback)
        if is_callback_async:
            await self._run_async(connection)
        else:
            await self._run_sync(connection, threads)

    async def _run_async(self, connection: AbstractRobustConnection) -> None:
        await self._run_async_callback(connection)

    async def _run_sync(
        self, connection: AbstractRobustConnection, threads: int
    ) -> None:
        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor(max_workers=threads) as executor:
            await self._run_sync_callback(connection, loop, executor)

    def _make_producer(
        self, connection: AbstractRobustConnection
    ) -> AsyncContextManager[Producer]:
        return make_producer(
            connection,
            self._downstream_pipe,
            size=self._pipe_length,
            durable=self._durable,
            backoff_seconds=self._backoff_seconds,
            max_retries=self._max_retries,
            exponential_backoff=self._exponential_backoff,
            max_backoff_seconds=self._max_backoff_seconds,
            apply_jitter=self._apply_jitter,
            persistent_messages=self._durable,
            clock=self._clock,
        )

    async def _run_async_callback(self, connection: AbstractRobustConnection) -> None:
        async with self._make_producer(connection) as producer:
            callback_generator = self._callback()
            if not isinstance(callback_generator, AsyncGenerator):
                raise TypeError(
                    f"Expected AsyncGenerator as callback, got: {type(callback_generator)}"
                )
            async for obj in callback_generator:
                if not isinstance(obj, pydantic.BaseModel):
                    raise TypeError(
                        f"Source callback must yield BaseModel instances, got {type(obj).__name__}"
                    )
                message = pydantic_model_to_message(obj)
                await producer.publish(message)

    async def _run_sync_callback(
        self,
        connection: AbstractRobustConnection,
        loop: AbstractEventLoop,
        executor: ThreadPoolExecutor,
    ) -> None:
        async with self._make_producer(connection) as producer:
            callback_generator = self._callback()
            if not isinstance(callback_generator, Generator):
                raise TypeError(
                    f"Expected Generator as callback, got: {type(callback_generator)}"
                )
            sentinel = object()
            while True:
                obj = await loop.run_in_executor(
                    executor, lambda: next(callback_generator, sentinel)
                )
                if obj is sentinel:
                    break
                if not isinstance(obj, pydantic.BaseModel):
                    raise TypeError(
                        f"Source callback must yield BaseModel instances, got {type(obj).__name__}"
                    )
                message = pydantic_model_to_message(obj)
                await producer.publish(message)


class Node(Runner):
    def __init__(
        self,
        name: str,
        upstream_pipe: str,
        downstream_pipe: str,
        callback: CallbackT,
        pipe_length: Optional[int] = None,
        durable: bool = False,
        backoff_seconds: float = 0.1,
        max_retries: Optional[int] = None,
        exponential_backoff: bool = False,
        max_backoff_seconds: Optional[float] = None,
        apply_jitter: bool = False,
        connection_max_retries: Optional[int] = None,
        connection_backoff_seconds: float = 1.0,
        prefetch_count: Optional[int] = None,
        clock: Clock = UtcClock(),
    ) -> None:
        self._name = name
        self._upstream_pipe = upstream_pipe
        self._downstream_pipe = downstream_pipe
        self._callback = callback
        self._pipe_length = pipe_length
        self._durable = durable
        self._backoff_seconds = backoff_seconds
        self._max_retries = max_retries
        self._exponential_backoff = exponential_backoff
        self._max_backoff_seconds = max_backoff_seconds
        self._apply_jitter = apply_jitter
        self._connection_max_retries = connection_max_retries
        self._connection_backoff_seconds = connection_backoff_seconds
        self._prefetch_count = prefetch_count
        self._clock = clock

    @property
    def name(self) -> str:
        return self._name

    async def run(self, connection: AbstractRobustConnection, threads: int) -> None:
        is_callback_async = _is_callback_async(self._callback)
        if is_callback_async:
            await self._run_async(connection)
        else:
            await self._run_sync(connection, threads)

    async def _run_async(self, connection: AbstractRobustConnection) -> None:
        while True:
            await self._run_async_callback(connection)

    async def _run_sync(
        self, connection: AbstractRobustConnection, threads: int
    ) -> None:
        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor(max_workers=threads) as executor:
            while True:
                await self._run_sync_callback(connection, loop, executor, threads)

    def _make_producer(
        self, connection: AbstractRobustConnection
    ) -> AsyncContextManager[Producer]:
        return make_producer(
            connection,
            self._downstream_pipe,
            size=self._pipe_length,
            durable=self._durable,
            backoff_seconds=self._backoff_seconds,
            max_retries=self._max_retries,
            exponential_backoff=self._exponential_backoff,
            max_backoff_seconds=self._max_backoff_seconds,
            apply_jitter=self._apply_jitter,
            persistent_messages=self._durable,
            clock=self._clock,
        )

    def _consume(
        self, connection: AbstractRobustConnection
    ) -> AsyncGenerator[AsyncContextManager[AbstractIncomingMessage], None]:
        return consume(
            connection,
            self._upstream_pipe,
            upstream_connection_max_retries=self._connection_max_retries,
            upstream_connection_backoff_seconds=self._connection_backoff_seconds,
            prefetch_count=self._prefetch_count,
            clock=self._clock,
        )

    async def _run_async_callback(self, connection: AbstractRobustConnection) -> None:
        async with self._make_producer(connection) as producer:
            async for message in self._consume(connection):
                async with message as current_message:
                    obj = message_to_pydantic_model(current_message, Bag)
                    awaitable = self._callback(obj)
                    if not isinstance(awaitable, Awaitable):
                        raise TypeError(
                            f"Expected callback to be awaitable, got {type(awaitable)}"
                        )
                    result = await awaitable
                    if result is not None:
                        await _handle_callback_output(result, producer)

    async def _run_sync_callback(
        self,
        connection: AbstractRobustConnection,
        loop: AbstractEventLoop,
        executor: ThreadPoolExecutor,
        threads: int,
    ) -> None:
        async with self._make_producer(connection) as producer:

            async def process(
                message_context: AsyncContextManager[AbstractIncomingMessage],
            ) -> None:
                async with message_context as current_message:
                    obj = message_to_pydantic_model(current_message, Bag)
                    if not isinstance(self._callback, Callable):
                        raise TypeError(
                            f"Expected callback to be callable, got {type(self._callback)}"
                        )
                    result = await loop.run_in_executor(executor, self._callback, obj)
                    if result is not None:
                        await _handle_callback_output(result, producer)

            await _dispatch_concurrent(self._consume(connection), threads, process)


class Sink(Runner):
    def __init__(
        self,
        name: str,
        upstream_pipe: str,
        callback: CallbackT,
        connection_max_retries: Optional[int] = None,
        connection_backoff_seconds: float = 1.0,
        prefetch_count: Optional[int] = None,
        clock: Clock = UtcClock(),
    ) -> None:
        self._name = name
        self._upstream_pipe = upstream_pipe
        self._callback = callback
        self._connection_max_retries = connection_max_retries
        self._connection_backoff_seconds = connection_backoff_seconds
        self._prefetch_count = prefetch_count
        self._clock = clock

    @property
    def name(self) -> str:
        return self._name

    async def run(self, connection: AbstractRobustConnection, threads: int) -> None:
        is_callback_async = _is_callback_async(self._callback)
        if is_callback_async:
            await self._run_async(connection)
        else:
            await self._run_sync(connection, threads)

    async def _run_async(self, connection: AbstractRobustConnection) -> None:
        while True:
            await self._run_async_callback(connection)

    async def _run_sync(
        self, connection: AbstractRobustConnection, threads: int
    ) -> None:
        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor(max_workers=threads) as executor:
            while True:
                await self._run_sync_callback(connection, loop, executor, threads)

    def _consume(
        self, connection: AbstractRobustConnection
    ) -> AsyncGenerator[AsyncContextManager[AbstractIncomingMessage], None]:
        return consume(
            connection,
            self._upstream_pipe,
            upstream_connection_max_retries=self._connection_max_retries,
            upstream_connection_backoff_seconds=self._connection_backoff_seconds,
            prefetch_count=self._prefetch_count,
            clock=self._clock,
        )

    async def _run_async_callback(self, connection: AbstractRobustConnection) -> None:
        async for message in self._consume(connection):
            async with message as current_message:
                obj = message_to_pydantic_model(current_message, Bag)
                awaitable = self._callback(obj)
                if not isinstance(awaitable, Awaitable):
                    raise TypeError(
                        f"Expected callback to be awaitable, got {type(awaitable).__name__}"
                    )
                await awaitable

    async def _run_sync_callback(
        self,
        connection: AbstractRobustConnection,
        loop: AbstractEventLoop,
        executor: ThreadPoolExecutor,
        threads: int,
    ) -> None:
        async def process(
            message_context: AsyncContextManager[AbstractIncomingMessage],
        ) -> None:
            async with message_context as current_message:
                obj = message_to_pydantic_model(current_message, Bag)
                await loop.run_in_executor(executor, self._callback, obj)

        await _dispatch_concurrent(self._consume(connection), threads, process)


class SinkKwargs(TypedDict, total=False):
    connection_max_retries: int
    connection_backoff_seconds: float
    prefetch_count: int


class SourceKwargs(TypedDict, total=False):
    pipe_length: int
    durable: bool
    backoff_seconds: float
    max_retries: int
    exponential_backoff: bool
    max_backoff_seconds: float
    apply_jitter: bool


class NodeKwargs(SourceKwargs, SinkKwargs):
    pass


RunnerKwargs = Union[SourceKwargs, SinkKwargs, NodeKwargs]


@dataclass(init=True, frozen=True)
class SourceDefinition:
    name: str
    callback: GeneratorCallbackT
    kwargs: SourceKwargs = field(default_factory=SourceKwargs)


@dataclass(init=True, frozen=True)
class NodeDefinition:
    name: str
    callback: CallbackT
    kwargs: SinkKwargs = field(default_factory=NodeKwargs)


class Pipeline:
    def __init__(self, runners: List[Runner]) -> None:
        self._runners = runners
        self._runners_map = {runner.name: runner for runner in runners}

    async def run(
        self, connection: AbstractRobustConnection, name: str, threads: int = 1
    ) -> None:
        runner = self._runners_map.get(name)
        if not runner:
            raise ValueError(f"Runner {name} not found in the pipeline")
        await runner.run(connection, threads)


class PipelineBuilder:
    def __init__(self, name: Optional[str] = None) -> None:
        self._name = name
        self._source_definition: Optional[SourceDefinition] = None
        self._node_definitions: List[NodeDefinition] = []
        self._sink_definition: Optional[NodeDefinition] = None

    def source(
        self, name: str, callback: GeneratorCallbackT, **kwargs: Unpack[SourceKwargs]
    ) -> Self:
        _validate_source_kwargs(**kwargs)
        full_name = _determine_full_node_name(name, self._name)
        self._source_definition = SourceDefinition(
            name=full_name, callback=callback, kwargs=SourceKwargs(**kwargs)
        )
        return self

    def node(
        self, name: str, callback: CallbackT, **kwargs: Unpack[NodeKwargs]
    ) -> Self:
        full_name = _determine_full_node_name(name, self._name)
        self._node_definitions.append(
            NodeDefinition(
                name=full_name, callback=callback, kwargs=NodeKwargs(**kwargs)
            )
        )
        return self

    def sink(
        self, name: str, callback: CallbackT, **kwargs: Unpack[SinkKwargs]
    ) -> Self:
        _validate_sink_kwargs(**kwargs)
        full_name = _determine_full_node_name(name, self._name)
        self._sink_definition = NodeDefinition(
            name=full_name, callback=callback, kwargs=SinkKwargs(**kwargs)
        )
        return self

    def build(self) -> Pipeline:
        if self._source_definition is None or self._sink_definition is None:
            raise PipelineDefinitionInvalid(
                "Pipeline must have at least source and sink"
            )
        node_definitions = list(self._node_definitions)
        runners: List[Runner] = []
        first_downstream = (
            node_definitions[0].name if node_definitions else self._sink_definition.name
        )
        runners.append(
            Source(
                name=self._source_definition.name,
                downstream_pipe=first_downstream,
                callback=self._source_definition.callback,
                **self._source_definition.kwargs,
            )
        )
        for i, node_def in enumerate(node_definitions):
            downstream = (
                node_definitions[i + 1]
                if i + 1 < len(node_definitions)
                else self._sink_definition
            )
            runners.append(
                Node(
                    name=node_def.name,
                    upstream_pipe=node_def.name,
                    downstream_pipe=downstream.name,
                    callback=node_def.callback,
                    **node_def.kwargs,
                )
            )
        runners.append(
            Sink(
                name=self._sink_definition.name,
                upstream_pipe=self._sink_definition.name,
                callback=self._sink_definition.callback,
                **self._sink_definition.kwargs,
            )
        )
        return Pipeline(runners)


def make_pipeline(name: Optional[str] = None) -> PipelineBuilder:
    return PipelineBuilder(name)


async def _dispatch_concurrent(
    consume_iter: AsyncGenerator[AsyncContextManager[AbstractIncomingMessage], None],
    threads: int,
    process_fn: Callable[
        [AsyncContextManager[AbstractIncomingMessage]], Awaitable[None]
    ],
) -> None:
    semaphore = asyncio.Semaphore(threads)
    pending = set()

    async def gated(
        message_context: AsyncContextManager[AbstractIncomingMessage],
    ) -> None:
        async with semaphore:
            await process_fn(message_context)

    async for message in consume_iter:
        task = asyncio.create_task(gated(message))
        pending.add(task)
        task.add_done_callback(pending.discard)

    if pending:
        await asyncio.gather(*pending)


async def _handle_callback_output(result: CallbackOutputT, producer: Producer) -> None:
    if _is_callback_result_iterable(result):
        results_serialized = [
            pydantic_model_to_message(item)
            for item in cast(Iterable[BaseModel], result)
        ]
        for item_serialized in results_serialized:
            await producer.publish(item_serialized)
    else:
        if not isinstance(result, BaseModel):
            raise TypeError(
                f"Callback must return a BaseModel instance, got {type(result).__name__}"
            )
        result_serialized = pydantic_model_to_message(result)
        await producer.publish(result_serialized)


def _is_callback_async(callback: Callable[..., ...]) -> bool:
    target = callback
    if not inspect.isfunction(callback) and not inspect.ismethod(callback):
        if hasattr(callback, "__call__"):
            target = callback.__call__
        else:
            return False

    return inspect.iscoroutinefunction(target) or inspect.isasyncgenfunction(target)


def _is_callback_result_iterable(
    value: CallbackOutputT,
) -> bool:
    return isinstance(value, Iterable) and not isinstance(value, BaseModel)


def _validate_source_kwargs(**kwargs: Unpack[NodeKwargs]) -> None:
    for kwarg_name, _ in kwargs.items():
        if kwarg_name not in get_type_hints(SourceKwargs):
            raise NodeKwargsInvalid(f"Invalid argument for source node: {kwarg_name}")


def _validate_sink_kwargs(**kwargs: Unpack[NodeKwargs]) -> None:
    for kwarg_name, _ in kwargs.items():
        if kwarg_name not in get_type_hints(SinkKwargs):
            raise NodeKwargsInvalid(f"Invalid argument for sink node: {kwarg_name}")


def _determine_full_node_name(
    node_name: str, pipeline_name: Optional[str] = None
) -> str:
    if pipeline_name is not None:
        return f"{pipeline_name}.{node_name}"
    else:
        return node_name



from typing import Union

import pydantic
from aio_pika import Message
from aio_pika.abc import AbstractIncomingMessage

from khnm.types import MessageObjectT


def pydantic_model_to_message(
    obj: pydantic.BaseModel,
) -> Message:
    bytes_ = obj.model_dump_json().encode("utf-8")
    return Message(bytes_)


def message_to_pydantic_model(
    message: Union[Message, AbstractIncomingMessage], model_class: type[MessageObjectT]
) -> MessageObjectT:
    decoded = message.body.decode("utf-8")
    obj = model_class.model_validate_json(decoded)
    return obj
